from __future__ import absolute_import

import base64
import json
import logging
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import pytz
from elasticsearch import Elasticsearch, helpers

from feast import Entity, FeatureView, RepoConfig
from feast.infra.key_encoding_utils import get_list_val_str, serialize_entity_key
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel


class ElasticSearchOnlineStoreConfig(FeastConfigBaseModel):
    """
    Configuration for the ElasticSearch online store.
    NOTE: The class *must* end with the `OnlineStoreConfig` suffix.
    """

    type: str = "elasticsearch"

    host: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None
    port: Optional[int] = None
    index: Optional[str] = None
    scheme: Optional[str] = "http"

    # The number of rows to write in a single batch
    write_batch_size: Optional[int] = 40

    # The length of the vector value
    vector_len: Optional[int] = 512

    # The vector similarity metric to use in KNN search
    # more details: https://www.elastic.co/guide/en/elasticsearch/reference/current/dense-vector.html
    similarity: Optional[str] = "cosine"


class ElasticSearchOnlineStore(OnlineStore):
    _client: Optional[Elasticsearch] = None

    def _get_client(self, config: RepoConfig) -> Elasticsearch:
        online_store_config = config.online_store
        assert isinstance(online_store_config, ElasticSearchOnlineStoreConfig)

        user = online_store_config.user if online_store_config.user is not None else ""
        password = (
            online_store_config.password
            if online_store_config.password is not None
            else ""
        )

        if self._client:
            return self._client
        else:
            self._client = Elasticsearch(
                hosts=[
                    {
                        "host": online_store_config.host or "localhost",
                        "port": online_store_config.port or 9200,
                        "scheme": online_store_config.scheme or "http",
                    }
                ],
                basic_auth=(user, password),
            )
            return self._client

    def _bulk_batch_actions(self, table: FeatureView, batch: List[Dict[str, Any]]):
        for row in batch:
            yield {
                "_index": table.name,
                "_id": f"{row['entity_key']}_{row['feature_name']}_{row['timestamp']}",
                "_source": row,
            }

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        insert_values = []
        for entity_key, values, timestamp, created_ts in data:
            entity_key_bin = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            )
            encoded_entity_key = base64.b64encode(entity_key_bin).decode("utf-8")
            timestamp = _to_naive_utc(timestamp)
            if created_ts is not None:
                created_ts = _to_naive_utc(created_ts)
            for feature_name, value in values.items():
                encoded_value = base64.b64encode(value.SerializeToString()).decode(
                    "utf-8"
                )
                vector_val = json.loads(get_list_val_str(value))
                insert_values.append(
                    {
                        "entity_key": encoded_entity_key,
                        "feature_name": feature_name,
                        "feature_value": encoded_value,
                        "timestamp": timestamp,
                        "created_ts": created_ts,
                        "vector_value": vector_val,
                    }
                )

        batch_size = config.online_store.write_batch_size
        for i in range(0, len(insert_values), batch_size):
            batch = insert_values[i : i + batch_size]
            actions = self._bulk_batch_actions(table, batch)
            helpers.bulk(self._get_client(config), actions)

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        if not requested_features:
            body = {
                "_source": {"excludes": ["vector_value"]},
                "query": {"match": {"entity_key": entity_keys}},
            }
        else:
            body = {
                "_source": {"excludes": ["vector_value"]},
                "query": {
                    "bool": {
                        "must": [
                            {"terms": {"entity_key": entity_keys}},
                            {"terms": {"feature_name": requested_features}},
                        ]
                    }
                },
            }
        response = self._get_client(config).search(index=table.name, body=body)
        results: List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]] = []
        for hit in response["hits"]["hits"]:
            results.append(
                (
                    hit["_source"]["timestamp"],
                    {hit["_source"]["feature_name"]: hit["_source"]["feature_value"]},
                )
            )
        return results

    def create_index(self, config: RepoConfig, table: FeatureView):
        """
        Create an index in ElasticSearch for the given table.
        TODO: This method can be exposed to users to customize the indexing functionality.
        Args:
            config: Feast repo configuration object.
            table: FeatureView table for which the index needs to be created.
        """
        index_mapping = {
            "properties": {
                "entity_key": {"type": "binary"},
                "feature_name": {"type": "keyword"},
                "feature_value": {"type": "binary"},
                "timestamp": {"type": "date"},
                "created_ts": {"type": "date"},
                "vector_value": {
                    "type": "dense_vector",
                    "dims": config.online_store.vector_len,
                    "index": "true",
                    "similarity": config.online_store.similarity,
                },
            }
        }
        self._get_client(config).indices.create(
            index=table.name, mappings=index_mapping
        )

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        # implement the update method
        for table in tables_to_delete:
            self._get_client(config).delete_by_query(index=table.name)
        for table in tables_to_keep:
            self.create_index(config, table)

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        project = config.project
        try:
            for table in tables:
                self._get_client(config).indices.delete(index=table.name)
        except Exception as e:
            logging.exception(f"Error deleting index in project {project}: {e}")
            raise

    def retrieve_online_documents(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_feature: str,
        embedding: List[float],
        top_k: int,
        *args,
        **kwargs,
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[ValueProto],
            Optional[ValueProto],
            Optional[ValueProto],
        ]
    ]:
        result: List[
            Tuple[
                Optional[datetime],
                Optional[ValueProto],
                Optional[ValueProto],
                Optional[ValueProto],
            ]
        ] = []
        response = self._get_client(config).search(
            index=table.name,
            knn={
                "field": "vector_value",
                "query_vector": embedding,
                "k": top_k,
            },
        )
        rows = response["hits"]["hits"][0:top_k]
        for row in rows:
            feature_value = row["_source"]["feature_value"]
            vector_value = row["_source"]["vector_value"]
            timestamp = row["_source"]["timestamp"]
            distance = row["_score"]
            timestamp = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")

            feature_value_proto = ValueProto()
            feature_value_proto.ParseFromString(base64.b64decode(feature_value))

            vector_value_proto = ValueProto(string_val=str(vector_value))
            distance_value_proto = ValueProto(float_val=distance)
            result.append(
                (
                    timestamp,
                    feature_value_proto,
                    vector_value_proto,
                    distance_value_proto,
                )
            )
        return result


def _to_naive_utc(ts: datetime):
    if ts.tzinfo is None:
        return ts
    else:
        return ts.astimezone(pytz.utc).replace(tzinfo=None)



import ast
import json
import logging
import traceback
from datetime import datetime
from typing import Any, Dict, List

import pyarrow as pa
import pyarrow.flight as fl

from feast import FeatureStore, FeatureView, utils
from feast.feature_logging import FeatureServiceLoggingSource
from feast.feature_view import DUMMY_ENTITY_NAME
from feast.infra.offline_stores.offline_utils import get_offline_store_from_config
from feast.saved_dataset import SavedDatasetStorage

logger = logging.getLogger(__name__)


class OfflineServer(fl.FlightServerBase):
    def __init__(self, store: FeatureStore, location: str, **kwargs):
        super(OfflineServer, self).__init__(location, **kwargs)
        self._location = location
        # A dictionary of configured flights, e.g. API calls received and not yet served
        self.flights: Dict[str, Any] = {}
        self.store = store
        self.offline_store = get_offline_store_from_config(store.config.offline_store)

    @classmethod
    def descriptor_to_key(self, descriptor: fl.FlightDescriptor):
        return (
            descriptor.descriptor_type.value,
            descriptor.command,
            tuple(descriptor.path or tuple()),
        )

    def _make_flight_info(self, key: Any, descriptor: fl.FlightDescriptor):
        endpoints = [fl.FlightEndpoint(repr(key), [self._location])]
        # TODO calculate actual schema from the given features
        schema = pa.schema([])

        return fl.FlightInfo(schema, descriptor, endpoints, -1, -1)

    def get_flight_info(
        self, context: fl.ServerCallContext, descriptor: fl.FlightDescriptor
    ):
        key = OfflineServer.descriptor_to_key(descriptor)
        if key in self.flights:
            return self._make_flight_info(key, descriptor)
        raise KeyError("Flight not found.")

    def list_flights(self, context: fl.ServerCallContext, criteria: bytes):
        for key, table in self.flights.items():
            if key[1] is not None:
                descriptor = fl.FlightDescriptor.for_command(key[1])
            else:
                descriptor = fl.FlightDescriptor.for_path(*key[2])

            yield self._make_flight_info(key, descriptor)

    # Expects to receive request parameters and stores them in the flights dictionary
    # Indexed by the unique command
    def do_put(
        self,
        context: fl.ServerCallContext,
        descriptor: fl.FlightDescriptor,
        reader: fl.MetadataRecordBatchReader,
        writer: fl.FlightMetadataWriter,
    ):
        key = OfflineServer.descriptor_to_key(descriptor)
        command = json.loads(key[1])
        if "api" in command:
            data = reader.read_all()
            logger.debug(f"do_put: command is{command}, data is {data}")
            self.flights[key] = data

            self._call_api(command, key)
        else:
            logger.warning(f"No 'api' field in command: {command}")

    def _call_api(self, command: dict, key: str):
        remove_data = False
        try:
            api = command["api"]
            if api == OfflineServer.offline_write_batch.__name__:
                self.offline_write_batch(command, key)
                remove_data = True
            elif api == OfflineServer.write_logged_features.__name__:
                self.write_logged_features(command, key)
                remove_data = True
            elif api == OfflineServer.persist.__name__:
                self.persist(command["retrieve_func"], command, key)
                remove_data = True
        except Exception as e:
            remove_data = True
            logger.exception(e)
            traceback.print_exc()
            raise e
        finally:
            if remove_data:
                # Get service is consumed, so we clear the corresponding flight and data
                del self.flights[key]

    def get_feature_view_by_name(
        self, fv_name: str, name_alias: str, project: str
    ) -> FeatureView:
        """
        Retrieves a feature view by name, including all subclasses of FeatureView.

        Args:
            fv_name: Name of feature view
            name_alias: Alias to be applied to the projection of the registered view
            project: Feast project that this feature view belongs to

        Returns:
            Returns either the specified feature view, or raises an exception if
            none is found
        """
        try:
            fv = self.store.registry.get_feature_view(name=fv_name, project=project)
            if name_alias is not None:
                for fs in self.store.registry.list_feature_services(project=project):
                    for p in fs.feature_view_projections:
                        if p.name_alias == name_alias:
                            logger.debug(
                                f"Found matching FeatureService {fs.name} with projection {p}"
                            )
                            fv = fv.with_projection(p)
            return fv
        except Exception:
            try:
                return self.store.registry.get_stream_feature_view(
                    name=fv_name, project=project
                )
            except Exception as e:
                logger.error(
                    f"Cannot find any FeatureView by name {fv_name} in project {project}"
                )
                raise e

    def list_feature_views_by_name(
        self, feature_view_names: List[str], name_aliases: List[str], project: str
    ) -> List[FeatureView]:
        return [
            remove_dummies(
                self.get_feature_view_by_name(
                    fv_name=fv_name, name_alias=name_aliases[index], project=project
                )
            )
            for index, fv_name in enumerate(feature_view_names)
        ]

    # Extracts the API parameters from the flights dictionary, delegates the execution to the FeatureStore instance
    # and returns the stream of data
    def do_get(self, context: fl.ServerCallContext, ticket: fl.Ticket):
        key = ast.literal_eval(ticket.ticket.decode())
        if key not in self.flights:
            logger.error(f"Unknown key {key}")
            return None

        command = json.loads(key[1])
        api = command["api"]
        logger.debug(f"get command is {command}")
        logger.debug(f"requested api is {api}")
        try:
            if api == OfflineServer.get_historical_features.__name__:
                table = self.get_historical_features(command, key).to_arrow()
            elif api == OfflineServer.pull_all_from_table_or_query.__name__:
                table = self.pull_all_from_table_or_query(command).to_arrow()
            elif api == OfflineServer.pull_latest_from_table_or_query.__name__:
                table = self.pull_latest_from_table_or_query(command).to_arrow()
            else:
                raise NotImplementedError
        except Exception as e:
            logger.exception(e)
            traceback.print_exc()
            raise e

        # Get service is consumed, so we clear the corresponding flight and data
        del self.flights[key]
        return fl.RecordBatchStream(table)

    def offline_write_batch(self, command: dict, key: str):
        feature_view_names = command["feature_view_names"]
        assert (
            len(feature_view_names) == 1
        ), "feature_view_names list should only have one item"
        name_aliases = command["name_aliases"]
        assert len(name_aliases) == 1, "name_aliases list should only have one item"
        project = self.store.config.project
        feature_views = self.list_feature_views_by_name(
            feature_view_names=feature_view_names,
            name_aliases=name_aliases,
            project=project,
        )

        assert len(feature_views) == 1
        table = self.flights[key]
        self.offline_store.offline_write_batch(
            self.store.config, feature_views[0], table, command["progress"]
        )

    def write_logged_features(self, command: dict, key: str):
        table = self.flights[key]
        feature_service = self.store.get_feature_service(
            command["feature_service_name"]
        )

        assert feature_service.logging_config is not None

        self.offline_store.write_logged_features(
            config=self.store.config,
            data=table,
            source=FeatureServiceLoggingSource(
                feature_service, self.store.config.project
            ),
            logging_config=feature_service.logging_config,
            registry=self.store.registry,
        )

    def pull_all_from_table_or_query(self, command: dict):
        return self.offline_store.pull_all_from_table_or_query(
            self.store.config,
            self.store.get_data_source(command["data_source_name"]),
            command["join_key_columns"],
            command["feature_name_columns"],
            command["timestamp_field"],
            utils.make_tzaware(datetime.fromisoformat(command["start_date"])),
            utils.make_tzaware(datetime.fromisoformat(command["end_date"])),
        )

    def pull_latest_from_table_or_query(self, command: dict):
        return self.offline_store.pull_latest_from_table_or_query(
            self.store.config,
            self.store.get_data_source(command["data_source_name"]),
            command["join_key_columns"],
            command["feature_name_columns"],
            command["timestamp_field"],
            command["created_timestamp_column"],
            utils.make_tzaware(datetime.fromisoformat(command["start_date"])),
            utils.make_tzaware(datetime.fromisoformat(command["end_date"])),
        )

    def list_actions(self, context):
        return [
            (
                OfflineServer.offline_write_batch.__name__,
                "Writes the specified arrow table to the data source underlying the specified feature view.",
            ),
            (
                OfflineServer.write_logged_features.__name__,
                "Writes logged features to a specified destination in the offline store.",
            ),
            (
                OfflineServer.persist.__name__,
                "Synchronously executes the underlying query and persists the result in the same offline store at the "
                "specified destination.",
            ),
        ]

    def get_historical_features(self, command: dict, key: str):
        # Extract parameters from the internal flights dictionary
        entity_df_value = self.flights[key]
        entity_df = pa.Table.to_pandas(entity_df_value)
        feature_view_names = command["feature_view_names"]
        name_aliases = command["name_aliases"]
        feature_refs = command["feature_refs"]
        project = command["project"]
        full_feature_names = command["full_feature_names"]
        feature_views = self.list_feature_views_by_name(
            feature_view_names=feature_view_names,
            name_aliases=name_aliases,
            project=project,
        )
        retJob = self.offline_store.get_historical_features(
            config=self.store.config,
            feature_views=feature_views,
            feature_refs=feature_refs,
            entity_df=entity_df,
            registry=self.store.registry,
            project=project,
            full_feature_names=full_feature_names,
        )
        return retJob

    def persist(self, retrieve_func: str, command: dict, key: str):
        try:
            if retrieve_func == OfflineServer.get_historical_features.__name__:
                ret_job = self.get_historical_features(command, key)
            elif (
                retrieve_func == OfflineServer.pull_latest_from_table_or_query.__name__
            ):
                ret_job = self.pull_latest_from_table_or_query(command)
            elif retrieve_func == OfflineServer.pull_all_from_table_or_query.__name__:
                ret_job = self.pull_all_from_table_or_query(command)
            else:
                raise NotImplementedError

            data_source = self.store.get_data_source(command["data_source_name"])
            storage = SavedDatasetStorage.from_data_source(data_source)
            ret_job.persist(storage, command["allow_overwrite"], command["timeout"])
        except Exception as e:
            logger.exception(e)
            traceback.print_exc()
            raise e

    def do_action(self, context: fl.ServerCallContext, action: fl.Action):
        pass

    def do_drop_dataset(self, dataset):
        pass


def remove_dummies(fv: FeatureView) -> FeatureView:
    """
    Removes dummmy IDs from FeatureView instances created with FeatureView.from_proto
    """
    if DUMMY_ENTITY_NAME in fv.entities:
        fv.entities = []
        fv.entity_columns = []
    return fv


def start_server(
    store: FeatureStore,
    host: str,
    port: int,
):
    location = "grpc+tcp://{}:{}".format(host, port)
    server = OfflineServer(store, location)
    logger.info(f"Offline store server serving on {location}")
    server.serve()



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple

import requests
from pydantic import StrictStr

from feast import Entity, FeatureView, RepoConfig
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel
from feast.type_map import python_values_to_proto_values
from feast.value_type import ValueType

logger = logging.getLogger(__name__)


class RemoteOnlineStoreConfig(FeastConfigBaseModel):
    """Remote Online store config for remote online store"""

    type: Literal["remote"] = "remote"
    """Online store type selector"""

    path: StrictStr = "http://localhost:6566"
    """ str: Path to metadata store.
    If type is 'remote', then this is a URL for registry server """


class RemoteOnlineStore(OnlineStore):
    """
    remote online store implementation wrapper to communicate with feast online server.
    """

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        raise NotImplementedError

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        assert isinstance(config.online_store, RemoteOnlineStoreConfig)
        config.online_store.__class__ = RemoteOnlineStoreConfig

        req_body = self._construct_online_read_api_json_request(
            entity_keys, table, requested_features
        )
        response = requests.post(
            f"{config.online_store.path}/get-online-features", data=req_body
        )
        if response.status_code == 200:
            logger.debug("Able to retrieve the online features from feature server.")
            response_json = json.loads(response.text)
            event_ts = self._get_event_ts(response_json)
            # Iterating over results and converting the API results in column format to row format.
            result_tuples: List[
                Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]
            ] = []
            for feature_value_index in range(len(entity_keys)):
                feature_values_dict: Dict[str, ValueProto] = dict()
                for index, feature_name in enumerate(
                    response_json["metadata"]["feature_names"]
                ):
                    if (
                        requested_features is not None
                        and feature_name in requested_features
                    ):
                        if (
                            response_json["results"][index]["statuses"][
                                feature_value_index
                            ]
                            == "PRESENT"
                        ):
                            message = python_values_to_proto_values(
                                [
                                    response_json["results"][index]["values"][
                                        feature_value_index
                                    ]
                                ],
                                ValueType.UNKNOWN,
                            )
                            feature_values_dict[feature_name] = message[0]
                        else:
                            feature_values_dict[feature_name] = ValueProto()
                result_tuples.append((event_ts, feature_values_dict))
            return result_tuples
        else:
            error_msg = f"Unable to retrieve the online store data using feature server API. Error_code={response.status_code}, error_message={response.reason}"
            logger.error(error_msg)
            raise RuntimeError(error_msg)

    def _construct_online_read_api_json_request(
        self,
        entity_keys: List[EntityKeyProto],
        table: FeatureView,
        requested_features: Optional[List[str]] = None,
    ) -> str:
        api_requested_features = []
        if requested_features is not None:
            for requested_feature in requested_features:
                api_requested_features.append(f"{table.name}:{requested_feature}")

        entity_values = []
        entity_key = ""
        for row in entity_keys:
            entity_key = row.join_keys[0]
            entity_values.append(
                getattr(row.entity_values[0], row.entity_values[0].WhichOneof("val"))
            )

        req_body = json.dumps(
            {
                "features": api_requested_features,
                "entities": {entity_key: entity_values},
            }
        )
        return req_body

    def _get_event_ts(self, response_json) -> datetime:
        event_ts = ""
        if len(response_json["results"]) > 1:
            event_ts = response_json["results"][1]["event_timestamps"][0]
        return datetime.fromisoformat(event_ts.replace("Z", "+00:00"))

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        pass

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        pass



import uuid
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import ibis
import ibis.selectors as s
import numpy as np
import pandas as pd
import pyarrow
from ibis.expr import datatypes as dt
from ibis.expr.types import Table
from pytz import utc

from feast.data_source import DataSource
from feast.feature_logging import LoggingConfig, LoggingSource
from feast.feature_view import FeatureView
from feast.infra.offline_stores import offline_utils
from feast.infra.offline_stores.file_source import (
    FileLoggingDestination,
)
from feast.infra.offline_stores.offline_store import (
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.offline_stores.offline_utils import (
    get_pyarrow_schema_from_batch_source,
)
from feast.infra.registry.base_registry import BaseRegistry
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage


def _get_entity_schema(entity_df: pd.DataFrame) -> Dict[str, np.dtype]:
    return dict(zip(entity_df.columns, entity_df.dtypes))


def pull_latest_from_table_or_query_ibis(
    config: RepoConfig,
    data_source: DataSource,
    join_key_columns: List[str],
    feature_name_columns: List[str],
    timestamp_field: str,
    created_timestamp_column: Optional[str],
    start_date: datetime,
    end_date: datetime,
    data_source_reader: Callable[[DataSource], Table],
    data_source_writer: Callable[[pyarrow.Table, DataSource], None],
    staging_location: Optional[str] = None,
    staging_location_endpoint_override: Optional[str] = None,
) -> RetrievalJob:
    fields = join_key_columns + feature_name_columns + [timestamp_field]
    if created_timestamp_column:
        fields.append(created_timestamp_column)
    start_date = start_date.astimezone(tz=utc)
    end_date = end_date.astimezone(tz=utc)

    table = data_source_reader(data_source)

    table = table.select(*fields)

    # TODO get rid of this fix
    if "__log_date" in table.columns:
        table = table.drop("__log_date")

    table = table.filter(
        ibis.and_(
            table[timestamp_field] >= ibis.literal(start_date),
            table[timestamp_field] <= ibis.literal(end_date),
        )
    )

    table = deduplicate(
        table=table,
        group_by_cols=join_key_columns,
        event_timestamp_col=timestamp_field,
        created_timestamp_col=created_timestamp_column,
    )

    return IbisRetrievalJob(
        table=table,
        on_demand_feature_views=[],
        full_feature_names=False,
        metadata=None,
        data_source_writer=data_source_writer,
        staging_location=staging_location,
        staging_location_endpoint_override=staging_location_endpoint_override,
    )


def _get_entity_df_event_timestamp_range(
    entity_df: pd.DataFrame, entity_df_event_timestamp_col: str
) -> Tuple[datetime, datetime]:
    entity_df_event_timestamp = entity_df.loc[
        :, entity_df_event_timestamp_col
    ].infer_objects()
    if pd.api.types.is_string_dtype(entity_df_event_timestamp):
        entity_df_event_timestamp = pd.to_datetime(entity_df_event_timestamp, utc=True)
    entity_df_event_timestamp_range = (
        entity_df_event_timestamp.min().to_pydatetime(),
        entity_df_event_timestamp.max().to_pydatetime(),
    )

    return entity_df_event_timestamp_range


def _to_utc(entity_df: pd.DataFrame, event_timestamp_col):
    entity_df_event_timestamp = entity_df.loc[:, event_timestamp_col].infer_objects()
    if pd.api.types.is_string_dtype(entity_df_event_timestamp):
        entity_df_event_timestamp = pd.to_datetime(entity_df_event_timestamp, utc=True)

    entity_df[event_timestamp_col] = entity_df_event_timestamp
    return entity_df


def _generate_row_id(
    entity_table: Table, feature_views: List[FeatureView], event_timestamp_col
) -> Table:
    all_entities = [event_timestamp_col]
    for fv in feature_views:
        if fv.projection.join_key_map:
            all_entities.extend(fv.projection.join_key_map.values())
        else:
            all_entities.extend([e.name for e in fv.entity_columns])

    r = ibis.literal("")

    for e in set(all_entities):
        r = r.concat(entity_table[e].cast("string"))  # type: ignore

    entity_table = entity_table.mutate(entity_row_id=r)

    return entity_table


def get_historical_features_ibis(
    config: RepoConfig,
    feature_views: List[FeatureView],
    feature_refs: List[str],
    entity_df: Union[pd.DataFrame, str],
    registry: BaseRegistry,
    project: str,
    data_source_reader: Callable[[DataSource], Table],
    data_source_writer: Callable[[pyarrow.Table, DataSource], None],
    full_feature_names: bool = False,
    staging_location: Optional[str] = None,
    staging_location_endpoint_override: Optional[str] = None,
) -> RetrievalJob:
    entity_schema = _get_entity_schema(
        entity_df=entity_df,
    )
    event_timestamp_col = offline_utils.infer_event_timestamp_from_entity_df(
        entity_schema=entity_schema,
    )

    # TODO get range with ibis
    timestamp_range = _get_entity_df_event_timestamp_range(
        entity_df, event_timestamp_col
    )

    entity_df = _to_utc(entity_df, event_timestamp_col)

    entity_table = ibis.memtable(entity_df)
    entity_table = _generate_row_id(entity_table, feature_views, event_timestamp_col)

    def read_fv(
        feature_view: FeatureView, feature_refs: List[str], full_feature_names: bool
    ) -> Tuple:
        fv_table: Table = data_source_reader(feature_view.batch_source)

        for old_name, new_name in feature_view.batch_source.field_mapping.items():
            if old_name in fv_table.columns:
                fv_table = fv_table.rename({new_name: old_name})

        timestamp_field = feature_view.batch_source.timestamp_field

        # TODO mutate only if tz-naive
        fv_table = fv_table.mutate(
            **{
                timestamp_field: fv_table[timestamp_field].cast(
                    dt.Timestamp(timezone="UTC")
                )
            }
        )

        full_name_prefix = feature_view.projection.name_alias or feature_view.name

        feature_refs = [
            fr.split(":")[1]
            for fr in feature_refs
            if fr.startswith(f"{full_name_prefix}:")
        ]

        if full_feature_names:
            fv_table = fv_table.rename(
                {f"{full_name_prefix}__{feature}": feature for feature in feature_refs}
            )

            feature_refs = [
                f"{full_name_prefix}__{feature}" for feature in feature_refs
            ]

        return (
            fv_table,
            feature_view.batch_source.timestamp_field,
            feature_view.batch_source.created_timestamp_column,
            feature_view.projection.join_key_map
            or {e.name: e.name for e in feature_view.entity_columns},
            feature_refs,
            feature_view.ttl,
        )

    res = point_in_time_join(
        entity_table=entity_table,
        feature_tables=[
            read_fv(feature_view, feature_refs, full_feature_names)
            for feature_view in feature_views
        ],
        event_timestamp_col=event_timestamp_col,
    )

    odfvs = OnDemandFeatureView.get_requested_odfvs(feature_refs, project, registry)

    substrait_odfvs = [fv for fv in odfvs if fv.mode == "substrait"]
    for odfv in substrait_odfvs:
        res = odfv.transform_ibis(res, full_feature_names)

    return IbisRetrievalJob(
        res,
        [fv for fv in odfvs if fv.mode != "substrait"],
        full_feature_names,
        metadata=RetrievalMetadata(
            features=feature_refs,
            keys=list(set(entity_df.columns) - {event_timestamp_col}),
            min_event_timestamp=timestamp_range[0],
            max_event_timestamp=timestamp_range[1],
        ),
        data_source_writer=data_source_writer,
        staging_location=staging_location,
        staging_location_endpoint_override=staging_location_endpoint_override,
    )


def pull_all_from_table_or_query_ibis(
    config: RepoConfig,
    data_source: DataSource,
    join_key_columns: List[str],
    feature_name_columns: List[str],
    timestamp_field: str,
    start_date: datetime,
    end_date: datetime,
    data_source_reader: Callable[[DataSource], Table],
    data_source_writer: Callable[[pyarrow.Table, DataSource], None],
    staging_location: Optional[str] = None,
    staging_location_endpoint_override: Optional[str] = None,
) -> RetrievalJob:
    fields = join_key_columns + feature_name_columns + [timestamp_field]
    start_date = start_date.astimezone(tz=utc)
    end_date = end_date.astimezone(tz=utc)

    table = data_source_reader(data_source)

    table = table.select(*fields)

    # TODO get rid of this fix
    if "__log_date" in table.columns:
        table = table.drop("__log_date")

    table = table.filter(
        ibis.and_(
            table[timestamp_field] >= ibis.literal(start_date),
            table[timestamp_field] <= ibis.literal(end_date),
        )
    )

    return IbisRetrievalJob(
        table=table,
        on_demand_feature_views=[],
        full_feature_names=False,
        metadata=None,
        data_source_writer=data_source_writer,
        staging_location=staging_location,
        staging_location_endpoint_override=staging_location_endpoint_override,
    )


def write_logged_features_ibis(
    config: RepoConfig,
    data: Union[pyarrow.Table, Path],
    source: LoggingSource,
    logging_config: LoggingConfig,
    registry: BaseRegistry,
):
    destination = logging_config.destination
    assert isinstance(destination, FileLoggingDestination)

    table = ibis.read_parquet(data) if isinstance(data, Path) else ibis.memtable(data)

    if destination.partition_by:
        kwargs = {"partition_by": destination.partition_by}
    else:
        kwargs = {}

    # TODO always write to directory
    table.to_parquet(f"{destination.path}/{uuid.uuid4().hex}-{{i}}.parquet", **kwargs)


def offline_write_batch_ibis(
    config: RepoConfig,
    feature_view: FeatureView,
    table: pyarrow.Table,
    progress: Optional[Callable[[int], Any]],
    data_source_writer: Callable[[pyarrow.Table, DataSource], None],
):
    pa_schema, column_names = get_pyarrow_schema_from_batch_source(
        config, feature_view.batch_source
    )
    if column_names != table.column_names:
        raise ValueError(
            f"The input pyarrow table has schema {table.schema} with the incorrect columns {table.column_names}. "
            f"The schema is expected to be {pa_schema} with the columns (in this exact order) to be {column_names}."
        )

    data_source_writer(ibis.memtable(table), feature_view.batch_source)


def deduplicate(
    table: Table,
    group_by_cols: List[str],
    event_timestamp_col: str,
    created_timestamp_col: Optional[str],
):
    order_by_fields = [ibis.desc(table[event_timestamp_col])]
    if created_timestamp_col:
        order_by_fields.append(ibis.desc(table[created_timestamp_col]))

    table = (
        table.group_by(by=group_by_cols)
        .order_by(order_by_fields)
        .mutate(rn=ibis.row_number())
    )

    return table.filter(table["rn"] == ibis.literal(0)).drop("rn")


def point_in_time_join(
    entity_table: Table,
    feature_tables: List[Tuple[Table, str, str, Dict[str, str], List[str], timedelta]],
    event_timestamp_col="event_timestamp",
):
    # TODO handle ttl
    all_entities = [event_timestamp_col]
    for (
        feature_table,
        timestamp_field,
        created_timestamp_field,
        join_key_map,
        _,
        _,
    ) in feature_tables:
        all_entities.extend(join_key_map.values())

    r = ibis.literal("")

    for e in set(all_entities):
        r = r.concat(entity_table[e].cast("string"))  # type: ignore

    entity_table = entity_table.mutate(entity_row_id=r)

    acc_table = entity_table

    for (
        feature_table,
        timestamp_field,
        created_timestamp_field,
        join_key_map,
        feature_refs,
        ttl,
    ) in feature_tables:
        predicates = [
            feature_table[k] == entity_table[v] for k, v in join_key_map.items()
        ]

        predicates.append(
            feature_table[timestamp_field] <= entity_table[event_timestamp_col],
        )

        if ttl:
            predicates.append(
                feature_table[timestamp_field]
                >= entity_table[event_timestamp_col] - ibis.literal(ttl)
            )

        feature_table = feature_table.inner_join(
            entity_table, predicates, lname="", rname="{name}_y"
        )

        feature_table = feature_table.drop(s.endswith("_y"))

        feature_table = deduplicate(
            table=feature_table,
            group_by_cols=["entity_row_id"],
            event_timestamp_col=timestamp_field,
            created_timestamp_col=created_timestamp_field,
        )

        select_cols = ["entity_row_id"]
        select_cols.extend(feature_refs)
        feature_table = feature_table.select(select_cols)

        acc_table = acc_table.left_join(
            feature_table,
            predicates=[feature_table.entity_row_id == acc_table.entity_row_id],
            lname="",
            rname="{name}_yyyy",
        )

        acc_table = acc_table.drop(s.endswith("_yyyy"))

    acc_table = acc_table.drop("entity_row_id")

    return acc_table


def list_s3_files(path: str, endpoint_url: str) -> List[str]:
    import boto3

    s3 = boto3.client("s3", endpoint_url=endpoint_url)
    if path.startswith("s3://"):
        path = path[len("s3://") :]
    bucket, prefix = path.split("/", 1)
    objects = s3.list_objects_v2(Bucket=bucket, Prefix=prefix)
    contents = objects["Contents"]
    files = [
        f"s3://{bucket}/{content['Key']}"
        for content in contents
        if content["Key"].endswith("parquet")
    ]
    return files


class IbisRetrievalJob(RetrievalJob):
    def __init__(
        self,
        table,
        on_demand_feature_views,
        full_feature_names,
        metadata,
        data_source_writer,
        staging_location,
        staging_location_endpoint_override,
    ) -> None:
        super().__init__()
        self.table = table
        self._on_demand_feature_views: List[OnDemandFeatureView] = (
            on_demand_feature_views
        )
        self._full_feature_names = full_feature_names
        self._metadata = metadata
        self.data_source_writer = data_source_writer
        self.staging_location = staging_location
        self.staging_location_endpoint_override = staging_location_endpoint_override

    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        return self.table.execute()

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pyarrow.Table:
        return self.table.to_pyarrow()

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: bool = False,
        timeout: Optional[int] = None,
    ):
        self.data_source_writer(
            self.table, storage.to_data_source(), "overwrite", allow_overwrite
        )

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata

    def supports_remote_storage_export(self) -> bool:
        return self.staging_location is not None

    def to_remote_storage(self) -> List[str]:
        path = self.staging_location + f"/{str(uuid.uuid4())}"

        storage_options = {"AWS_ENDPOINT_URL": self.staging_location_endpoint_override}

        self.table.to_delta(path, storage_options=storage_options)

        return list_s3_files(path, self.staging_location_endpoint_override)



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import os
import sqlite3
import struct
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple, Union

import sqlite_vec
from google.protobuf.internal.containers import RepeatedScalarFieldContainer
from pydantic import StrictStr

from feast import Entity
from feast.feature_view import FeatureView
from feast.infra.infra_object import SQLITE_INFRA_OBJECT_CLASS_TYPE, InfraObject
from feast.infra.key_encoding_utils import serialize_entity_key
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.core.InfraObject_pb2 import InfraObject as InfraObjectProto
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.protos.feast.core.SqliteTable_pb2 import SqliteTable as SqliteTableProto
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import FloatList as FloatListProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.utils import to_naive_utc


class SqliteOnlineStoreConfig(FeastConfigBaseModel):
    """Online store config for local (SQLite-based) store"""

    type: Literal["sqlite", "feast.infra.online_stores.sqlite.SqliteOnlineStore"] = (
        "sqlite"
    )
    """ Online store type selector"""

    path: StrictStr = "data/online.db"
    """ (optional) Path to sqlite db """

    vec_enabled: Optional[bool] = False
    """ (optional) Enable or disable sqlite-vss for vector search"""

    vector_len: Optional[int] = 512
    """ (optional) Length of the vector to be stored in the database"""


class SqliteOnlineStore(OnlineStore):
    """
    SQLite implementation of the online store interface. Not recommended for production usage.

    Attributes:
        _conn: SQLite connection.
    """

    _conn: Optional[sqlite3.Connection] = None

    @staticmethod
    def _get_db_path(config: RepoConfig) -> str:
        assert (
            config.online_store.type == "sqlite"
            or config.online_store.type.endswith("SqliteOnlineStore")
        )

        if config.repo_path and not Path(config.online_store.path).is_absolute():
            db_path = str(config.repo_path / config.online_store.path)
        else:
            db_path = config.online_store.path
        return db_path

    def _get_conn(self, config: RepoConfig):
        if not self._conn:
            db_path = self._get_db_path(config)
            self._conn = _initialize_conn(db_path)
            if sys.version_info[0:2] == (3, 10):
                self._conn.enable_load_extension(True)  # type: ignore
                sqlite_vec.load(self._conn)

        return self._conn

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[
                EntityKeyProto,
                Dict[str, ValueProto],
                datetime,
                Optional[datetime],
            ]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        conn = self._get_conn(config)

        project = config.project

        with conn:
            for entity_key, values, timestamp, created_ts in data:
                entity_key_bin = serialize_entity_key(
                    entity_key,
                    entity_key_serialization_version=config.entity_key_serialization_version,
                )
                timestamp = to_naive_utc(timestamp)
                if created_ts is not None:
                    created_ts = to_naive_utc(created_ts)

                table_name = _table_id(project, table)
                for feature_name, val in values.items():
                    if config.online_store.vec_enabled:
                        vector_bin = serialize_f32(
                            val.float_list_val.val, config.online_store.vector_len
                        )  # type: ignore
                        conn.execute(
                            f"""
                                    UPDATE {table_name}
                                    SET value = ?, vector_value = ?, event_ts = ?, created_ts = ?
                                    WHERE (entity_key = ? AND feature_name = ?)
                                """,
                            (
                                # SET
                                val.SerializeToString(),
                                vector_bin,
                                timestamp,
                                created_ts,
                                # WHERE
                                entity_key_bin,
                                feature_name,
                            ),
                        )

                        conn.execute(
                            f"""INSERT OR IGNORE INTO {table_name}
                                (entity_key, feature_name, value, vector_value, event_ts, created_ts)
                                VALUES (?, ?, ?, ?, ?, ?)""",
                            (
                                entity_key_bin,
                                feature_name,
                                val.SerializeToString(),
                                vector_bin,
                                timestamp,
                                created_ts,
                            ),
                        )

                    else:
                        conn.execute(
                            f"""
                                UPDATE {table_name}
                                SET value = ?, event_ts = ?, created_ts = ?
                                WHERE (entity_key = ? AND feature_name = ?)
                            """,
                            (
                                # SET
                                val.SerializeToString(),
                                timestamp,
                                created_ts,
                                # WHERE
                                entity_key_bin,
                                feature_name,
                            ),
                        )

                        conn.execute(
                            f"""INSERT OR IGNORE INTO {table_name}
                                (entity_key, feature_name, value, event_ts, created_ts)
                                VALUES (?, ?, ?, ?, ?)""",
                            (
                                entity_key_bin,
                                feature_name,
                                val.SerializeToString(),
                                timestamp,
                                created_ts,
                            ),
                        )
                if progress:
                    progress(1)

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        conn = self._get_conn(config)
        cur = conn.cursor()

        result: List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]] = []

        # Fetch all entities in one go
        cur.execute(
            f"SELECT entity_key, feature_name, value, event_ts "
            f"FROM {_table_id(config.project, table)} "
            f"WHERE entity_key IN ({','.join('?' * len(entity_keys))}) "
            f"ORDER BY entity_key",
            [
                serialize_entity_key(
                    entity_key,
                    entity_key_serialization_version=config.entity_key_serialization_version,
                )
                for entity_key in entity_keys
            ],
        )
        rows = cur.fetchall()

        rows = {
            k: list(group) for k, group in itertools.groupby(rows, key=lambda r: r[0])
        }
        for entity_key in entity_keys:
            entity_key_bin = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            )
            res = {}
            res_ts = None
            for _, feature_name, val_bin, ts in rows.get(entity_key_bin, []):
                val = ValueProto()
                val.ParseFromString(val_bin)
                res[feature_name] = val
                res_ts = ts

            if not res:
                result.append((None, None))
            else:
                result.append((res_ts, res))
        return result

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        conn = self._get_conn(config)
        project = config.project

        for table in tables_to_keep:
            conn.execute(
                f"CREATE TABLE IF NOT EXISTS {_table_id(project, table)} (entity_key BLOB, feature_name TEXT, value BLOB, vector_value BLOB, event_ts timestamp, created_ts timestamp,  PRIMARY KEY(entity_key, feature_name))"
            )
            conn.execute(
                f"CREATE INDEX IF NOT EXISTS {_table_id(project, table)}_ek ON {_table_id(project, table)} (entity_key);"
            )

        for table in tables_to_delete:
            conn.execute(f"DROP TABLE IF EXISTS {_table_id(project, table)}")

    def plan(
        self, config: RepoConfig, desired_registry_proto: RegistryProto
    ) -> List[InfraObject]:
        project = config.project

        infra_objects: List[InfraObject] = [
            SqliteTable(
                path=self._get_db_path(config),
                name=_table_id(project, FeatureView.from_proto(view)),
            )
            for view in [
                *desired_registry_proto.feature_views,
                *desired_registry_proto.stream_feature_views,
            ]
        ]
        return infra_objects

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        try:
            os.unlink(self._get_db_path(config))
        except FileNotFoundError:
            pass

    def retrieve_online_documents(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_feature: str,
        embedding: List[float],
        top_k: int,
        distance_metric: Optional[str] = None,
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[ValueProto],
            Optional[ValueProto],
            Optional[ValueProto],
        ]
    ]:
        """

        Args:
            config: Feast configuration object
            table: FeatureView object as the table to search
            requested_feature: The requested feature as the column to search
            embedding: The query embedding to search for
            top_k: The number of items to return
        Returns:
            List of tuples containing the event timestamp, the document feature, the vector value, and the distance
        """
        project = config.project

        if not config.online_store.vec_enabled:
            raise ValueError("sqlite-vss is not enabled in the online store config")

        conn = self._get_conn(config)
        cur = conn.cursor()

        # Convert the embedding to a binary format instead of using SerializeToString()
        query_embedding_bin = serialize_f32(embedding, config.online_store.vector_len)
        table_name = _table_id(project, table)

        cur.execute(
            f"""
            CREATE VIRTUAL TABLE vec_example using vec0(
                vector_value float[{config.online_store.vector_len}]
        );
        """
        )

        # Currently I can only insert the embedding value without crashing SQLite, will report a bug
        cur.execute(
            f"""
            INSERT INTO vec_example(rowid, vector_value)
            select rowid, vector_value from {table_name}
        """
        )
        cur.execute(
            """
            INSERT INTO vec_example(rowid, vector_value)
                VALUES (?, ?)
        """,
            (0, query_embedding_bin),
        )

        # Have to join this with the {table_name} to get the feature name and entity_key
        # Also the `top_k` doesn't appear to be working for some reason
        cur.execute(
            f"""
            select
                fv.entity_key,
                f.vector_value,
                fv.value,
                f.distance,
                fv.event_ts
            from (
                select
                    rowid,
                    vector_value,
                    distance
                from vec_example
                where vector_value match ?
                order by distance
                limit ?
            ) f
            left join {table_name} fv
            on f.rowid = fv.rowid
        """,
            (query_embedding_bin, top_k),
        )

        rows = cur.fetchall()

        result: List[
            Tuple[
                Optional[datetime],
                Optional[ValueProto],
                Optional[ValueProto],
                Optional[ValueProto],
            ]
        ] = []

        for entity_key, _, string_value, distance, event_ts in rows:
            feature_value_proto = ValueProto()
            feature_value_proto.ParseFromString(string_value if string_value else b"")
            vector_value_proto = ValueProto(
                float_list_val=FloatListProto(val=embedding)
            )
            distance_value_proto = ValueProto(float_val=distance)

            result.append(
                (
                    event_ts,
                    feature_value_proto,
                    vector_value_proto,
                    distance_value_proto,
                )
            )

        return result


def _initialize_conn(db_path: str):
    Path(db_path).parent.mkdir(exist_ok=True)
    return sqlite3.connect(
        db_path,
        detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
        check_same_thread=False,
    )


def _table_id(project: str, table: FeatureView) -> str:
    return f"{project}_{table.name}"


def serialize_f32(
    vector: Union[RepeatedScalarFieldContainer[float], List[float]], vector_length: int
) -> bytes:
    """serializes a list of floats into a compact "raw bytes" format"""
    return struct.pack(f"{vector_length}f", *vector)


def deserialize_f32(byte_vector: bytes, vector_length: int) -> List[float]:
    """deserializes a list of floats from a compact "raw bytes" format"""
    num_floats = vector_length // 4  # 4 bytes per float
    return list(struct.unpack(f"{num_floats}f", byte_vector))


class SqliteTable(InfraObject):
    """
    A Sqlite table managed by Feast.

    Attributes:
        path: The absolute path of the Sqlite file.
        name: The name of the table.
        conn: SQLite connection.
    """

    path: str
    conn: sqlite3.Connection

    def __init__(self, path: str, name: str):
        super().__init__(name)
        self.path = path
        self.conn = _initialize_conn(path)

    def to_infra_object_proto(self) -> InfraObjectProto:
        sqlite_table_proto = self.to_proto()
        return InfraObjectProto(
            infra_object_class_type=SQLITE_INFRA_OBJECT_CLASS_TYPE,
            sqlite_table=sqlite_table_proto,
        )

    def to_proto(self) -> Any:
        sqlite_table_proto = SqliteTableProto()
        sqlite_table_proto.path = self.path
        sqlite_table_proto.name = self.name
        return sqlite_table_proto

    @staticmethod
    def from_infra_object_proto(infra_object_proto: InfraObjectProto) -> Any:
        return SqliteTable(
            path=infra_object_proto.sqlite_table.path,
            name=infra_object_proto.sqlite_table.name,
        )

    @staticmethod
    def from_proto(sqlite_table_proto: SqliteTableProto) -> Any:
        return SqliteTable(
            path=sqlite_table_proto.path,
            name=sqlite_table_proto.name,
        )

    def update(self):
        if sys.version_info[0:2] == (3, 10):
            self.conn.enable_load_extension(True)
            sqlite_vec.load(self.conn)
        self.conn.execute(
            f"CREATE TABLE IF NOT EXISTS {self.name} (entity_key BLOB, feature_name TEXT, value BLOB, vector_value BLOB, event_ts timestamp, created_ts timestamp,  PRIMARY KEY(entity_key, feature_name))"
        )
        self.conn.execute(
            f"CREATE INDEX IF NOT EXISTS {self.name}_ek ON {self.name} (entity_key);"
        )

    def teardown(self):
        self.conn.execute(f"DROP TABLE IF EXISTS {self.name}")
